"""
# Secret

A package dealing with:
    - retrieving and resolving a secret from a 1Password vault
"""

__all__ = [
    "init_client",
    "resolve"
]

from .client import init_client
from .resolver import resolve



"""
# Secret Reference

A package dealing with:
    - expanding local environment variables within a secret reference URI
"""

__all__ = [
    "expand_env_vars"
]

from .expander import expand_env_vars



"""
Token - Module concerning retrieving a token value for a 1Password Service
Account
"""

from typing import Callable

from .. import shell_command

_TOKEN_ENV_VAR_NAME: str = "OP_SERVICE_ACCOUNT_TOKEN"
_POWERSHELL_TOKEN_ENV_VAR_NAME: str = f"$ENV:{_TOKEN_ENV_VAR_NAME}"
_SHELL_TOKEN_ENV_VAR_NAME: str = f"${_TOKEN_ENV_VAR_NAME}"

def get_token(
    platform: str,
    shell_command_resolver: Callable[[str], str]
) -> str:
    """
    Returns token from the local environment and errors if it is empty.
    """
    token_env_var_name: str
    if platform == "Windows":
        token_env_var_name = _POWERSHELL_TOKEN_ENV_VAR_NAME
    else:
        token_env_var_name = _SHELL_TOKEN_ENV_VAR_NAME

    token: str = shell_command.run(shell_command_resolver, token_env_var_name)

    if not token:
        raise ValueError(f"No value found for {token_env_var_name}")

    return token



"""
# Shell Command

A package dealing with:
    - resolve the platform-appropriate command to fetch environment variables
"""

__all__ = [
    "resolve",
    "run"
]

from .resolver import resolve
from .runner import run



"""
Resolver - a module for resolving the platform-appropriate command to fetch
environment variables.
"""

import os
from typing import Callable


_DEFAULT_SHELL: str = "bash"
_POWERSHELL_COMMAND: Callable[[str], str] = lambda env_var: (
    "powershell -command "
    f"\"$ExecutionContext.InvokeCommand.ExpandString({env_var})\""
)
# NOTE: Using an interactive mode command (bash/zsh/fish -ic) seemed to be
# the only way to access a user's env vars on a Mac outside Plover's
# environment.
_SHELL_COMMAND: Callable[[str], Callable[[str], str]] = lambda shell: (
    lambda env_var: f"{shell} -ic 'echo {env_var}'"
)

def resolve(platform: str) -> Callable[[str], str]:
    """
    Resolves a shell command for a given platform.
    """
    if platform == "Windows":
        return _POWERSHELL_COMMAND

    return _SHELL_COMMAND(
        os.getenv("SHELL", _DEFAULT_SHELL).split("/")[-1]
    )



"""
Plover entry point extension module for Plover 1Password

    - https://plover.readthedocs.io/en/latest/plugin-dev/extensions.html
    - https://plover.readthedocs.io/en/latest/plugin-dev/meta.html
"""

import asyncio
import platform
from typing import Callable

from plover.engine import StenoEngine
from plover.formatting import (
    _Action,
    _Context
)
from plover.registry import registry

from onepassword.client import Client

from . import (
    secret,
    secret_reference,
    service_account,
    shell_command
)


class OnePassword:
    """
    Extension class that also registers a meta plugin.
    The meta deals with retrieving secrets from 1Password
    """

    _client: Client
    _engine: StenoEngine
    _platform: str
    _shell_command: Callable[[str], str]

    def __init__(self, engine: StenoEngine) -> None:
        self._engine = engine

    def start(self) -> None:
        """
        Sets up the meta plugin and service account token.
        """
        self._platform = platform.system()
        self._shell_command = shell_command.resolve(self._platform)
        service_account_token = service_account.get_token(
            self._platform,
            self._shell_command
        )
        # The SDK client "sets up an authenticated session with the 1Password
        # servers and automatically refreshes it whenever it expires", so it
        # should hopefully be okay to locally cache a single client instance
        # here to handle all connections rather than create a new one on every
        # connection.
        # REF: https://developer.1password.com/docs/sdks/concepts/#state-management
        self._client = asyncio.run(
            secret.init_client(service_account_token)
        )
        registry.register_plugin(
            "meta",
            "1PASSWORD",
            lambda ctx, argument : asyncio.run(
                self._one_password(ctx, argument)
            )
        )

    def stop(self) -> None:
        """
        Stops the plugin -- no custom action needed.
        """

    async def _one_password(self, ctx: _Context, argument: str) -> _Action:
        """
        Retrieves a secret from 1Password based on the secret reference passed
        in as an argument in the steno outline, and outputs it.
        """
        op_secret_reference: str = secret_reference.expand_env_vars(
            self._shell_command,
            argument
        )
        secret_value: str = await secret.resolve(
            self._client,
            op_secret_reference
        )

        action: _Action = ctx.new_action()
        action.text = secret_value
        return action



"""
Module to resolve a given 1Password secret reference URI to a secret contained
in a vault.
"""

from onepassword.client import Client

from . import error


async def resolve(client: Client, secret_reference: str) -> str:
    """
    Resolves a single secret from a secret reference URI.
    """
    if not secret_reference:
        raise ValueError("Secret Reference cannot be blank")

    try:
        secret: str = await client.secrets.resolve(secret_reference)
    except Exception as exc: # pylint: disable=broad-except
        error.handle_ffi_error(exc, secret_reference)
        raise ValueError(exc) from exc

    return secret



"""
Expander - a module for dealing with expansion of ENV vars in a secret
reference URI.
"""

from typing import Callable

from .. import shell_command

_ENV_VAR_SYNTAX: str = "$"

def expand_env_vars(
    shell_command_resolver: Callable[[str], str],
    secret_reference: str
) -> str:
    """
    Expands env vars in a secret reference. Returns immediately if no env vars
    contained in secret reference string.
    """
    if _ENV_VAR_SYNTAX not in secret_reference:
        return secret_reference

    return shell_command.run(shell_command_resolver, secret_reference)



"""
Runner - a module for running shell commands.
"""

import subprocess
from typing import Callable


def run(shell_command_resolver: Callable[[str], str], target: str) -> str:
    """
    Runs a provided shell command against target in a subprocess.
    """
    command: str = shell_command_resolver(target)
    result: str = subprocess.run(
        command,
        capture_output=True,
        check=False,
        encoding="utf-8",
        shell=True
    ).stdout.strip()

    return result



"""
Module to initialise a 1Password client
"""

from onepassword.client import Client

from ..__version__ import __version__


_INTEGRATION_NAME: str = "Plover 1Password plugin integration"

async def init_client(service_account_token: str) -> Client:
    """
    Initialises a 1Password client to retrieve secrets.
    """
    return (
        await Client.authenticate(
            auth=service_account_token,
            integration_name=_INTEGRATION_NAME,
            integration_version=__version__
        )
    )



"""
Module to handle errors raised from 1Password's uniffi_core C shared libraries
but not handled in the Python SDK.
See: https://github.com/1Password/onepassword-sdk-python/tree/main/src/onepassword/lib
"""

_SERVICE_ACCOUNT_TOKEN_INVALID_ERROR: str = (
    "invalid service account token, please make sure you provide a valid "
    "service account token as parameter: service account deserialization "
    "failed, please create another token"
)
_SERVICE_ACCOUNT_TOKEN_INVALID_FORMAT_ERROR: str = (
    "invalid user input: encountered the following errors: "
    "service account token had invalid format"
)
_SECRET_REFERENCE_INVALID_FORMAT_ERROR: str = (
    "error resolving secret reference: "
    "secret reference has invalid format - "
    "must be \"op://<vault>/<item>/[section/]field\""
)
_SECRET_REFERENCE_MISSING_PREFIX_ERROR: str = (
    "error resolving secret reference: "
    "secret reference is not prefixed with \"op://\""
)
_SECRET_REFERENCE_VAULT_NOT_FOUND_ERROR: str = (
    "error resolving secret reference: "
    "no vault matched the secret reference query"
)
_SECRET_REFERENCE_ITEM_NOT_FOUND_ERROR: str = (
    "error resolving secret reference: "
    "no item matched the secret reference query"
)
_SECRET_REFERENCE_SECTION_NOT_FOUND_ERROR: str = (
    "error resolving secret reference: "
    "no section matched the secret reference query"
)
_SECRET_REFERENCE_FIELD_NOT_FOUND_ERROR: str = (
    "error resolving secret reference: "
    "the specified field cannot be found within the item"
)

def handle_ffi_error(exc: Exception, secret_reference: str) -> None:
    """
    Handles errors generated from 1Password's ffi libraries and re-raises them
    with more plugin-relevant messages.
    """
    error_message: str = str(exc)

    if _SERVICE_ACCOUNT_TOKEN_INVALID_ERROR in error_message:
        raise ValueError(
            "Service Account Token is invalid. "
            "Create another token and restart Plover."
        ) from exc

    if _SERVICE_ACCOUNT_TOKEN_INVALID_FORMAT_ERROR in error_message:
        raise ValueError(
            "Service Account Token has invalid format. "
            "Fix token format or create a new one and restart Plover."
        ) from exc

    if _SECRET_REFERENCE_INVALID_FORMAT_ERROR in error_message:
        raise ValueError(
            "Secret Reference has invalid format. "
            "URI must be \"op://<vault>/<item>/[section/]field\". "
            f"You provided {secret_reference}."
        ) from exc

    if _SECRET_REFERENCE_MISSING_PREFIX_ERROR in error_message:
        raise ValueError(
            "Secret Reference needs to be prefixed with \"op://\". "
            f"You provided {secret_reference}."
        ) from exc

    if _SECRET_REFERENCE_VAULT_NOT_FOUND_ERROR in error_message:
        raise ValueError(
            "Vault specified not found in Secret Reference "
            f"{secret_reference}."
        ) from exc

    if _SECRET_REFERENCE_ITEM_NOT_FOUND_ERROR in error_message:
        raise ValueError(
            "Item specified not found in Secret Reference "
            f"{secret_reference}."
        ) from exc

    if _SECRET_REFERENCE_SECTION_NOT_FOUND_ERROR in error_message:
        raise ValueError(
            "Section specified not found in Secret Reference "
            f"{secret_reference}."
        ) from exc

    if _SECRET_REFERENCE_FIELD_NOT_FOUND_ERROR in error_message:
        raise ValueError(
            "Field specified not found in Secret Reference "
            f"{secret_reference}."
        ) from exc
